/*
  stack-a.c
  Stack Allocator.
  Written by Alex Verstak.
  License: use at your own risk.
*/

#include <assert.h>
#include <stdio.h>
#include "stack-a.h"
#include "macros.h"
                                       /*defines*/
#ifndef ALLOCATION_GRANULARITY
#define ALLOCATION_GRANULARITY 0
#endif

#define stacka_shift (12+ALLOCATION_GRANULARITY)
#define stacka_add   ((1<<stacka_shift)-1)

void *stacka_offset=0;                 /*pool start*/
uint *stacka_start=0;                  /*stack start*/
uint  stacka_top=0;                    /*stack top*/
uint  stacka_sz=0;                     /*stack size*/

/*
  Get stack size.
  Pre:  none.
  Post: returns stack size.
  Note: max pool size is 4G-4K.
*/
int stacka_size(uint sz)
{
  return (sz>>stacka_shift)*sizeof(uint);
}

/*
  Initialize pool.
  Pre:  t is array of sufficient size;
        off is pool offset;
        sz is pool size.
  Post: returns 1 if initialized or 0 if error.
*/
int stacka_init(uint *t, void* off, uint sz)
{
  if((uint)off+sz<=(uint)off||!t)
    return 0;
  stacka_offset=off; stacka_start=t;
  stacka_sz=sz>>stacka_shift;
  stacka_alloc_all();
  return 1;
}

/*
  Allocate all pool.
  Pre:  none.
  Post: none.
*/
void stacka_alloc_all()
{
  assert(stacka_start&&stacka_sz&&stacka_top<=stacka_sz);
  stacka_top=stacka_sz;
}

/*
  Free all pool.
  Pre:  none.
  Post: none.
*/
void stacka_free_all()
{
  int i;
  assert(stacka_start&&stacka_sz&&stacka_top<=stacka_sz);
  stacka_top=stacka_sz;
  for(i=0;i<stacka_sz;i++)
    stacka_start[--stacka_top]=(uint)stacka_offset+i*(1<<stacka_shift);
}

/*
  Allocate page.
  Pre:  sz is page size.
  Post: pointer returned.
*/
void *stacka_alloc(uint sz)
{
  assert(stacka_start&&stacka_sz&&stacka_top<=stacka_sz);
  if(sz>(1<<stacka_shift)||stacka_top>=stacka_sz)
    return 0;
  return (void*)(stacka_start[stacka_top++]);
}

/*
  Free page.
  Pre:  p points to page.
  Post: 1 returned if page freed or 0 otherwise.
  Note: it doesn't check if the page is already free!!!
*/
int stacka_free(void *p)
{
  assert(stacka_start&&stacka_sz&&stacka_top<=stacka_sz);
  if((uint)p<(uint)stacka_offset||(uint)p>=(uint)stacka_offset+stacka_sz*(1<<stacka_shift))
    return 0;
  stacka_start[--stacka_top]=(uint)p;
  return 1;
}

/*
  Free pages at absolute address.
  Pre:  p is address, sz is size.
  Post: 1 is returned if ok, 0 otherwise.
  Note: it doesn't check if the pages are already free!!!
*/
int stacka_free_at(void *p,uint sz)
{
  uint i;
  assert(stacka_start&&stacka_sz&&stacka_top<=stacka_sz);
  sz=sz>>stacka_shift;
  if(!sz||(uint)p<(uint)stacka_offset||\
     (uint)p+(sz<<stacka_shift)>(uint)stacka_offset+stacka_sz*(1<<stacka_shift))
    return 0;
  for(i=0;i<sz;i++)
    stacka_free((void*)((int)p+i*(1<<stacka_shift)));
  return 1;
}

void stacka_stat()
{
  if(!stacka_start||!stacka_sz||stacka_top>stacka_sz)
    printf("Stack allocator not initialized.\n");
  else
  {
    printf("Stack allocator initialized.\n");
    printf("Address %x, %d blocks (%dKB), ",stacka_offset,stacka_sz,stacka_sz<<stacka_shift-10);
    printf("%d blocks free (%dKB).\n",stacka_sz-stacka_top,stacka_sz-stacka_top<<stacka_shift-10);
  }
}

/*------------------- end of stack-a.c -------------------*/

